/*******************************************************************************\
 * Copyright (c) 1996, 1997 by Markus Koelblin
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
\*******************************************************************************/

/*******************************************************************************\
 * xcolmap.c                                                                   *
 *                                                                             *
 * Implements a simple colormap monitor for PseudoColor visuals using          *
 * X-Toolkit and Xaw-Widgets                                                   *
 *                                                                             *
 * Author:  Markus Koelblin (Markus.Koelblin@hadiko.de) 1996, 1997             *
\*******************************************************************************/

const char version[]= "xcolmap version 1.0";

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/AsciiText.h>

#include <stdio.h>
#include "xcolmap.xbm"

/* Macro constants */
#define CLASS_NAME     "Xcolmap"

#define COLORS_ROW     16
#define COLORS_WIDTH   256
#define COLORS_MAX     256


/* Global colormap application data */
typedef struct tagAppData
{
    Display    *display;
    Window     map_win;
    GC         map_gc;
    Widget     topLevel;
    Widget     text;

    Dimension  height;
    Dimension  width;
    unsigned   cell_size;
}
AppData;

static Atom      wm_delete_window, wm_protocols;
static Pixmap    icon;
static AppData   app;


/*******************************************************************************\
 *  ButtonPressEV()                                                            *
 *  Acts on button press event for colorForm widget                            *
\*******************************************************************************/

void ButtonPressEV(Widget w, XtPointer client_data, XEvent *event, Boolean* cont)
{
    const int cell = ((AppData*)client_data)->cell_size;
    const int col  = event->xbutton.x / cell;
    const int row  = event->xbutton.y / cell;
    XColor color;

    /* Get pixel index from button down coordinates */
    color.pixel = col + row * COLORS_ROW;

    if (color.pixel < COLORS_MAX)
    {
	Display *display  = ((AppData*)client_data)->display;
	Colormap colormap = DefaultColormap(display, DefaultScreen(display));
	char     strbuf[16];

	XQueryColor(display, colormap, &color);
	sprintf(strbuf, "%3lu: #%02x%02x%02x",
		color.pixel,
		color.red  /256,
		color.green/256,
		color.blue /256);
	XtVaSetValues(((AppData*)client_data)->text, XtNstring, strbuf, NULL);
    } 
}

/******************************************************************************\
 *  RedrawAC() action                                                         *
 *  this draws the colormap                                                   *
\******************************************************************************/

void RedrawAC(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    Display *const display = app.display;
    const Window   window  = app.map_win;
    const GC       draw_gc = app.map_gc;
    const unsigned cell    = app.cell_size;
    unsigned       x       = 0;
    unsigned       y       = 0;
    Pixel          i;

    XSetForeground(display, draw_gc, (Pixel) 0);
    XFillRectangle(display, window, draw_gc, 0, 0, app.width, app.height);

    /*
     * Display COLORS_MAX entries of the colormap
     */
    for (i = 0; i < COLORS_MAX; i++)
    {
	XSetForeground(display, draw_gc, i);
	XFillRectangle(display, window, draw_gc, x, y, cell, cell);
	x+= cell;
	if (((i+1) % COLORS_ROW) == 0)
	{
	    x = 0;
	    y+= cell;
	}   
    }
}

/*******************************************************************************\
 *  ResizeAC() action                                                          *
 *  things to do, when the window is resized                                   *
\*******************************************************************************/

void ResizeAC(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    Arg args[2];

    /* Get new size of the window */
    XtSetArg(args[0], XtNheight, &app.height);
    XtSetArg(args[1], XtNwidth,  &app.width);
    XtGetValues(w, args, 2);

    /* Calculate new cell size */
    app.cell_size = (app.width < app.height) ? (app.width >> 4) : (app.height >> 4);

    /* Redraw colormap */
    RedrawAC(w, NULL, NULL, NULL);
}

/******************************************************************************\
 *  QuitAC()  action handler                                                  *
\******************************************************************************/

void QuitAC(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    if(event->type == ClientMessage && event->xclient.type == wm_protocols)
	if (event->xclient.data.l[0] != wm_delete_window)
	    return;

    XFreePixmap(app.display, icon);
    XFreeGC(app.display, app.map_gc);
    exit(0);
}

/******************************************************************************\
 *  QuitCB()  callback                                                        *
\******************************************************************************/

void QuitCB(Widget w, XtPointer client_data, XtPointer call_data)
{
    XFreePixmap(app.display, icon);
    XFreeGC(app.display, app.map_gc);
    exit(0);
}


/******************************************************************************/

static String fallback_resources[] =
{
    "*textEdit*background: lightyellow",
    "*textEdit.foreground: black",
    "*quit.label: Quit",
    NULL,
};

static XtActionsRec window_actions[] = 
{
    /* Defines the actions list */
    {"quit_action", QuitAC  },
    {"redraw_form", RedrawAC},
    {"resize_form", ResizeAC}
};

static String wm_trans = "<ClientMessage>WM_PROTOCOLS: quit_action()\n";


/******************************************************************************\
 *  main()                                                                    *
\******************************************************************************/

int main(int argc, char **argv)
{
    Widget         mainForm, colorForm, quit;
    XtTranslations xt_trans;
    XtAppContext   app_context;

    /* Setup standard size for the window */
    app.height    = COLORS_WIDTH;
    app.width     = COLORS_WIDTH;
    app.cell_size = COLORS_WIDTH / COLORS_ROW;

    /* initialize Toolkit and widget tree */
    app.topLevel = XtVaAppInitialize(&app_context, CLASS_NAME, NULL, 0, &argc, argv,
				     fallback_resources, NULL);

    mainForm  = XtVaCreateManagedWidget("mainForm", formWidgetClass, app.topLevel,
					XtNright,  XtChainRight,
					XtNleft,   XtChainLeft,
					XtNtop,    XtChainTop,
					XtNbottom, XtChainTop,
					NULL);

    xt_trans  = XtParseTranslationTable("<Configure>: resize_form()\n<Expose>: redraw_form()");
    colorForm = XtVaCreateManagedWidget("colorForm", formWidgetClass, mainForm,
					XtNtranslations, xt_trans,
					XtNheight, app.height,
					XtNwidth,  app.width,
					XtNright,  XtChainRight,
					XtNleft,   XtChainLeft,
					XtNbottom, XtChainBottom,
					XtNtop,    XtChainTop,
					NULL);

    quit      = XtVaCreateManagedWidget("quit", commandWidgetClass, mainForm,
					XtNfromVert, colorForm,
					XtNbottom, XtChainBottom,
					XtNtop, XtChainBottom,
					NULL);

    xt_trans  = XtParseTranslationTable("#override\n <Key>Return: no-op()\n\
                                         <Key>Linefeed: no-op()\nCtrl<Key>M: no-op()\n\
                                         Ctrl<Key>J: no-op()\nCtrl<Key>s: no-op()\n");

    app.text  =  XtVaCreateManagedWidget("textEdit", asciiTextWidgetClass, mainForm,
					 XtNfromHoriz, quit,
					 XtNfromVert, colorForm,
					 XtNbottom, XtChainBottom,
					 XtNtop, XtChainBottom,
					 XtNeditType, XawtextEdit,
					 XtNwrap, XawtextWrapNever,
					 XtNresize, XawtextResizeWidth,
					 XtNtranslations, xt_trans,
					 XtNwidth,  app.width/2,
					 XtNstring, "",
					 NULL);

    /* Add actions callbacks and event handler */
    XtAppAddActions(app_context, window_actions, XtNumber(window_actions));
    XtAddEventHandler(colorForm, ButtonPressMask, 0, ButtonPressEV, (XtPointer) &app);
    XtAddCallback(quit, XtNcallback, QuitCB, NULL);
    XtRealizeWidget(app.topLevel);

    /* Setup rest of application data */
    app.display = XtDisplay(colorForm);
    app.map_win = XtWindow(colorForm);
    app.map_gc  = XCreateGC(app.display, app.map_win, 0, NULL);

    /* Setup WM_DELETE_WINDOW action handler */
    wm_protocols     = XInternAtom(app.display, "WM_PROTOCOLS", False);
    wm_delete_window = XInternAtom(app.display, "WM_DELETE_WINDOW", False);
    XtOverrideTranslations(app.topLevel, XtParseTranslationTable(wm_trans));
    XSetWMProtocols(app.display, XtWindow(app.topLevel), &wm_delete_window, 1);

    /* Create and install application icon */
    icon = XCreateBitmapFromData(app.display, XtWindow(app.topLevel), (char*)xcolmap_bits,
				 xcolmap_width, xcolmap_height);
    XtVaSetValues(app.topLevel, XtNiconPixmap, icon, NULL);

    /* Enter the event loop */
    XtAppMainLoop(app_context);

    /* Never reached */
    return 0;
}

/*************************************************************************/
